# Copyright 2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import List
import unittest

from kfp import dsl
from kfp.dsl import component_factory
from kfp.dsl import Input
from kfp.dsl import Output
from kfp.dsl import structures
from kfp.dsl.component_decorator import component
from kfp.dsl.types.artifact_types import Artifact
from kfp.dsl.types.artifact_types import Model
from kfp.dsl.types.type_annotations import OutputPath


def strip_kfp_version(command: List[str]) -> List[str]:
    return [
        re.sub(r"'kfp==(\d+).(\d+).(\d+)(-[a-z]+.\d+)?'", 'kfp', c)
        for c in command
    ]


class TestGetPackagesToInstallCommand(unittest.TestCase):

    def test_with_no_user_packages_to_install(self):
        packages_to_install = []

        command = component_factory._get_packages_to_install_command(
            packages_to_install=packages_to_install)
        self.assertEqual(
            strip_kfp_version(command),
            strip_kfp_version([
                'sh', '-c',
                '\nif ! [ -x "$(command -v pip)" ]; then\n    python3 -m ensurepip || python3 -m ensurepip --user || apt-get install python3-pip\nfi\n\nPIP_DISABLE_PIP_VERSION_CHECK=1 python3 -m pip install --quiet --no-warn-script-location \'kfp==2.1.2\' \'--no-deps\' \'typing-extensions>=3.7.4,<5; python_version<"3.9"\' && "$0" "$@"\n'
            ]))

    def test_with_no_user_packages_to_install_and_install_kfp_false(self):
        packages_to_install = []

        command = component_factory._get_packages_to_install_command(
            packages_to_install=packages_to_install,
            install_kfp_package=False,
        )
        self.assertEqual(command, [])

    def test_with_no_user_packages_to_install_and_kfp_package_path(self):
        packages_to_install = []

        command = component_factory._get_packages_to_install_command(
            packages_to_install=packages_to_install,
            kfp_package_path='git+https://github.com/kubeflow/pipelines.git@master#subdirectory=sdk/python'
        )

        self.assertEqual(
            strip_kfp_version(command),
            strip_kfp_version([
                'sh', '-c',
                '\nif ! [ -x "$(command -v pip)" ]; then\n    python3 -m ensurepip || python3 -m ensurepip --user || apt-get install python3-pip\nfi\n\nPIP_DISABLE_PIP_VERSION_CHECK=1 python3 -m pip install --quiet --no-warn-script-location \'git+https://github.com/kubeflow/pipelines.git@master#subdirectory=sdk/python\' && "$0" "$@"\n'
            ]))

    def test_with_no_user_packages_to_install_and_kfp_package_path_and_install_kfp_false(
            self):
        packages_to_install = []

        command = component_factory._get_packages_to_install_command(
            packages_to_install=packages_to_install,
            kfp_package_path='git+https://github.com/kubeflow/pipelines.git@master#subdirectory=sdk/python',
            install_kfp_package=False,
        )
        self.assertEqual(command, [])

    def test_with_user_packages_to_install_and_kfp_package_path_and_install_kfp_false(
            self):
        packages_to_install = ['sklearn']

        command = component_factory._get_packages_to_install_command(
            packages_to_install=packages_to_install,
            kfp_package_path='git+https://github.com/kubeflow/pipelines.git@master#subdirectory=sdk/python',
            install_kfp_package=False,
        )

        self.assertEqual(
            strip_kfp_version(command),
            strip_kfp_version([
                'sh', '-c',
                '\nif ! [ -x "$(command -v pip)" ]; then\n    python3 -m ensurepip || python3 -m ensurepip --user || apt-get install python3-pip\nfi\n\nPIP_DISABLE_PIP_VERSION_CHECK=1 python3 -m pip install --quiet --no-warn-script-location \'sklearn\' && "$0" "$@"\n'
            ]))

    def test_with_no_user_packages_to_install_and_kfp_package_path_and_target_image(
            self):
        packages_to_install = []

        command = component_factory._get_packages_to_install_command(
            packages_to_install=packages_to_install,
            target_image='gcr.io/my-kfp-image',
            kfp_package_path='./sdk/python')

        self.assertEqual(command, [])

    def test_with_no_user_packages_to_install_and_kfp_package_path_and_target_image_and_install_kfp_false(
            self):
        packages_to_install = []

        command = component_factory._get_packages_to_install_command(
            packages_to_install=packages_to_install,
            target_image='gcr.io/my-kfp-image',
            kfp_package_path='./sdk/python',
            install_kfp_package=False)

        self.assertEqual(command, [])

    def test_with_user_packages_to_install_and_no_pip_index_url(self):
        packages_to_install = ['package1', 'package2']

        command = component_factory._get_packages_to_install_command(
            packages_to_install=packages_to_install)

        self.assertEqual(
            strip_kfp_version(command),
            strip_kfp_version([
                'sh', '-c', '\n'
                'if ! [ -x "$(command -v pip)" ]; then\n'
                '    python3 -m ensurepip || python3 -m ensurepip --user || apt-get install '
                'python3-pip\n'
                'fi\n'
                '\n'
                'PIP_DISABLE_PIP_VERSION_CHECK=1 python3 -m pip install --quiet '
                "--no-warn-script-location 'package1' 'package2'  &&  python3 -m pip install "
                "--quiet --no-warn-script-location kfp '--no-deps' "
                '\'typing-extensions>=3.7.4,<5; python_version<"3.9"\' && "$0" "$@"\n'
            ]))

    def test_with_packages_to_install_with_pip_index_url(self):
        packages_to_install = ['package1', 'package2']
        pip_index_urls = ['https://myurl.org/simple']

        command = component_factory._get_packages_to_install_command(
            packages_to_install=packages_to_install,
            pip_index_urls=pip_index_urls,
        )

        self.assertEqual(
            strip_kfp_version(command),
            strip_kfp_version([
                'sh', '-c', '\n'
                'if ! [ -x "$(command -v pip)" ]; then\n'
                '    python3 -m ensurepip || python3 -m ensurepip --user || apt-get install '
                'python3-pip\n'
                'fi\n'
                '\n'
                'PIP_DISABLE_PIP_VERSION_CHECK=1 python3 -m pip install --quiet '
                '--no-warn-script-location --index-url https://myurl.org/simple '
                "--trusted-host https://myurl.org/simple 'package1' 'package2'  &&  python3 "
                '-m pip install --quiet --no-warn-script-location --index-url '
                'https://myurl.org/simple --trusted-host https://myurl.org/simple kfp '
                '\'--no-deps\' \'typing-extensions>=3.7.4,<5; python_version<"3.9"\' && "$0" '
                '"$@"\n'
            ]))

    def test_with_packages_to_install_with_pip_index_url_and_trusted_host(self):
        packages_to_install = ['package1', 'package2']
        pip_index_urls = ['https://myurl.org/simple']
        pip_trusted_hosts = ['myurl.org']

        command = component_factory._get_packages_to_install_command(
            packages_to_install=packages_to_install,
            pip_index_urls=pip_index_urls,
            pip_trusted_hosts=pip_trusted_hosts,
        )

        self.assertEqual(
            strip_kfp_version(command),
            strip_kfp_version([
                'sh', '-c', '\n'
                'if ! [ -x "$(command -v pip)" ]; then\n'
                '    python3 -m ensurepip || python3 -m ensurepip --user || apt-get install '
                'python3-pip\n'
                'fi\n'
                '\n'
                'PIP_DISABLE_PIP_VERSION_CHECK=1 python3 -m pip install --quiet '
                '--no-warn-script-location --index-url https://myurl.org/simple '
                "--trusted-host myurl.org 'package1' 'package2'  &&  python3 -m pip install "
                '--quiet --no-warn-script-location --index-url https://myurl.org/simple '
                "--trusted-host myurl.org kfp '--no-deps' 'typing-extensions>=3.7.4,<5; "
                'python_version<"3.9"\' && "$0" "$@"\n'
            ]))

    def test_with_packages_to_install_with_pip_index_url_and_empty_trusted_host(
            self):
        packages_to_install = ['package1', 'package2']
        pip_index_urls = ['https://myurl.org/simple']

        command = component_factory._get_packages_to_install_command(
            packages_to_install=packages_to_install,
            pip_index_urls=pip_index_urls,
            pip_trusted_hosts=[],
        )

        self.assertEqual(
            strip_kfp_version(command),
            strip_kfp_version([
                'sh', '-c', '\n'
                'if ! [ -x "$(command -v pip)" ]; then\n'
                '    python3 -m ensurepip || python3 -m ensurepip --user || apt-get install '
                'python3-pip\n'
                'fi\n'
                '\n'
                'PIP_DISABLE_PIP_VERSION_CHECK=1 python3 -m pip install --quiet '
                "--no-warn-script-location --index-url https://myurl.org/simple 'package1' "
                "'package2'  &&  python3 -m pip install --quiet --no-warn-script-location "
                "--index-url https://myurl.org/simple kfp '--no-deps' "
                '\'typing-extensions>=3.7.4,<5; python_version<"3.9"\' && "$0" "$@"\n'
            ]))


class TestInvalidParameterName(unittest.TestCase):

    def test_output_named_Output(self):

        with self.assertRaisesRegex(ValueError,
                                    r'"Output" is an invalid parameter name.'):

            @component
            def comp(Output: OutputPath(str)):
                pass

    def test_output_named_Output_with_string_output(self):

        with self.assertRaisesRegex(ValueError,
                                    r'"Output" is an invalid parameter name.'):

            @component
            def comp(Output: OutputPath(str), text: str) -> str:
                pass


class TestExtractComponentInterfaceListofArtifacts(unittest.TestCase):

    def test_python_component_input(self):

        def comp(i: Input[List[Model]]):
            ...

        component_spec = component_factory.extract_component_interface(comp)
        self.assertEqual(component_spec.name, 'comp')
        self.assertEqual(component_spec.description, None)
        self.assertEqual(
            component_spec.inputs, {
                'i':
                    structures.InputSpec(
                        type='system.Model@0.0.1',
                        default=None,
                        is_artifact_list=True)
            })

    def test_custom_container_component_input(self):

        def comp(i: Input[List[Artifact]]):
            ...

        component_spec = component_factory.extract_component_interface(
            comp, containerized=True)
        self.assertEqual(component_spec.name, 'comp')
        self.assertEqual(component_spec.description, None)
        self.assertEqual(
            component_spec.inputs, {
                'i':
                    structures.InputSpec(
                        type='system.Artifact@0.0.1',
                        default=None,
                        is_artifact_list=True)
            })

    def test_pipeline_input(self):

        def comp(i: Input[List[Model]]):
            ...

        component_spec = component_factory.extract_component_interface(comp)
        self.assertEqual(component_spec.name, 'comp')
        self.assertEqual(component_spec.description, None)
        self.assertEqual(
            component_spec.inputs, {
                'i':
                    structures.InputSpec(
                        type='system.Model@0.0.1',
                        default=None,
                        is_artifact_list=True)
            })


class TestArtifactStringInInputpathOutputpath(unittest.TestCase):

    def test_unknown(self):

        @dsl.component
        def comp(
                i: dsl.InputPath('MyCustomType'),
                o: dsl.OutputPath('MyCustomType'),
        ):
            ...

        self.assertEqual(comp.component_spec.outputs['o'].type,
                         'system.Artifact@0.0.1')
        self.assertFalse(comp.component_spec.outputs['o'].is_artifact_list)
        self.assertEqual(comp.component_spec.inputs['i'].type,
                         'system.Artifact@0.0.1')
        self.assertFalse(comp.component_spec.inputs['i'].is_artifact_list)

    def test_known_v1_back_compat(self):

        @dsl.component
        def comp(
                i: dsl.InputPath('Dataset'),
                o: dsl.OutputPath('Dataset'),
        ):
            ...

        self.assertEqual(comp.component_spec.outputs['o'].type,
                         'system.Dataset@0.0.1')
        self.assertFalse(comp.component_spec.outputs['o'].is_artifact_list)
        self.assertEqual(comp.component_spec.inputs['i'].type,
                         'system.Dataset@0.0.1')
        self.assertFalse(comp.component_spec.inputs['i'].is_artifact_list)


class TestOutputListsOfArtifactsTemporarilyBlocked(unittest.TestCase):

    def test_python_component(self):
        with self.assertRaisesRegex(
                ValueError,
                r"Output lists of artifacts are only supported for pipelines\. Got output list of artifacts for output parameter 'output_list' of component 'comp'\."
        ):

            @dsl.component
            def comp(output_list: Output[List[Artifact]]):
                ...

    def test_container_component(self):
        with self.assertRaisesRegex(
                ValueError,
                r"Output lists of artifacts are only supported for pipelines\. Got output list of artifacts for output parameter 'output_list' of component 'comp'\."
        ):

            @dsl.container_component
            def comp(output_list: Output[List[Artifact]]):
                return dsl.ContainerSpec(image='alpine')


class TestPythonEOLWarning(unittest.TestCase):

    def test_default_base_image(self):

        with self.assertWarnsRegex(
                FutureWarning,
                r"The default base_image used by the @dsl\.component decorator will switch from 'python:3\.9' to 'python:3\.10' on Oct 1, 2025\. To ensure your existing components work with versions of the KFP SDK released after that date, you should provide an explicit base_image argument and ensure your component works as intended on Python 3\.10\."
        ):

            @dsl.component
            def foo():
                pass


if __name__ == '__main__':
    unittest.main()



# Copyright 2023 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for config.py."""
import os
import unittest
from unittest import mock

from kfp import local
from kfp.local import config


class LocalRunnerConfigTest(unittest.TestCase):

    def setUp(self):
        config.LocalExecutionConfig.instance = None

    def test_local_runner_config_init(self):
        """Test instance attributes with one constructor call."""
        config.LocalExecutionConfig(
            pipeline_root='my/local/root',
            runner=local.SubprocessRunner(use_venv=True),
            raise_on_error=True,
        )

        instance = config.LocalExecutionConfig.instance

        self.assertEqual(instance.pipeline_root, 'my/local/root')
        self.assertEqual(instance.runner, local.SubprocessRunner(use_venv=True))
        self.assertIs(instance.raise_on_error, True)

    def test_local_runner_config_is_singleton(self):
        """Test instance attributes with multiple constructor calls."""
        config.LocalExecutionConfig(
            pipeline_root='my/local/root',
            runner=local.SubprocessRunner(),
            raise_on_error=True,
        )
        config.LocalExecutionConfig(
            pipeline_root='other/local/root',
            runner=local.SubprocessRunner(use_venv=False),
            raise_on_error=False,
        )

        instance = config.LocalExecutionConfig.instance

        self.assertEqual(instance.pipeline_root, 'other/local/root')
        self.assertEqual(instance.runner,
                         local.SubprocessRunner(use_venv=False))
        self.assertFalse(instance.raise_on_error, False)

    def test_validate_success(self):
        config.LocalExecutionConfig(
            pipeline_root='other/local/root',
            runner=local.SubprocessRunner(use_venv=False),
            raise_on_error=False,
        )
        config.LocalExecutionConfig.validate()

    def test_validate_fail(self):
        with self.assertRaisesRegex(
                RuntimeError,
                r"Local environment not initialized. Please run 'kfp\.local\.init\(\)' before executing tasks locally\."
        ):
            config.LocalExecutionConfig.validate()


class TestInitCalls(unittest.TestCase):

    def setUp(self):
        config.LocalExecutionConfig.instance = None

    def test_init_more_than_once(self):
        """Tests config instance attributes with one init() call."""
        local.init(
            pipeline_root='my/local/root',
            runner=local.SubprocessRunner(use_venv=True),
        )

        instance = config.LocalExecutionConfig.instance

        self.assertEqual(instance.pipeline_root, 'my/local/root')
        self.assertEqual(instance.runner, local.SubprocessRunner(use_venv=True))

    def test_init_more_than_once(self):
        """Test config instance attributes with multiple init() calls."""
        local.init(
            pipeline_root='my/local/root',
            runner=local.SubprocessRunner(),
        )
        local.init(
            pipeline_root='other/local/root',
            runner=local.SubprocessRunner(use_venv=False),
            raise_on_error=False,
        )

        instance = config.LocalExecutionConfig.instance

        self.assertEqual(instance.pipeline_root,
                         os.path.abspath('other/local/root'))
        self.assertEqual(instance.runner,
                         local.SubprocessRunner(use_venv=False))
        self.assertFalse(instance.raise_on_error, False)

    def test_runner_validation(self):
        """Test config instance attributes with multiple init() calls."""
        with self.assertRaisesRegex(
                ValueError,
                r'Got unknown runner foo of type str\. Runner should be one of the following types: SubprocessRunner\.'
        ):
            local.init(runner='foo')


class TestDockerRunner(unittest.TestCase):

    def test_import_error(self):
        with mock.patch.dict('sys.modules', {'docker': None}):
            with self.assertRaisesRegex(
                    ImportError,
                    r"Package 'docker' must be installed to use 'DockerRunner'\. Install it using 'pip install docker'\."
            ):
                local.DockerRunner()

    def test_good_container_args(self):
        local.DockerRunner(network_mode='none')

    def test_unknown_container_args(self):
        with self.assertRaisesRegex(
                ValueError,
                r'Unsupported `docker run` arguments, see Package `docker` for details: .*'
        ):
            local.DockerRunner(image='spaghetti', favorite_vegetable='zucchini')

    def test_excess_container_args(self):
        with self.assertRaisesRegex(
                ValueError,
                r'The following docker run arguments should not be specififed: .*'
        ):
            local.DockerRunner(image='spaghetti')


if __name__ == '__main__':
    unittest.main()



# Copyright 2024 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Assorted utilities."""

from google.protobuf import json_format
from google.protobuf import struct_pb2
from kfp.pipeline_spec import pipeline_spec_pb2


def struct_to_executor_spec(
    struct: struct_pb2.Struct,
) -> pipeline_spec_pb2.PipelineDeploymentConfig.ExecutorSpec:
    executor_spec = pipeline_spec_pb2.PipelineDeploymentConfig.ExecutorSpec()
    json_format.ParseDict(json_format.MessageToDict(struct), executor_spec)
    return executor_spec



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.components.types.artifact_types."""

import contextlib
import json
import os
import unittest

from absl.testing import parameterized
from kfp import dsl
from kfp.dsl.types import artifact_types


class ArtifactsTest(unittest.TestCase):

    def test_complex_metrics(self):
        metrics = dsl.ClassificationMetrics()
        metrics.log_roc_data_point(threshold=0.1, tpr=98.2, fpr=96.2)
        metrics.log_roc_data_point(threshold=24.3, tpr=24.5, fpr=98.4)
        metrics.set_confusion_matrix_categories(['dog', 'cat', 'horses'])
        metrics.log_confusion_matrix_row('dog', [2, 6, 0])
        metrics.log_confusion_matrix_cell('cat', 'dog', 3)
        metrics.log_confusion_matrix_cell('horses', 'horses', 3)
        metrics.metadata['test'] = 1.0
        with open(
                os.path.join(
                    os.path.dirname(__file__), 'test_data',
                    'expected_io_types_classification_metrics.json')
        ) as json_file:
            expected_json = json.load(json_file)
            self.assertEqual(expected_json, metrics.metadata)

    def test_complex_metrics_bulk_loading(self):
        metrics = dsl.ClassificationMetrics()
        metrics.log_roc_curve(
            fpr=[85.1, 85.1, 85.1],
            tpr=[52.6, 52.6, 52.6],
            threshold=[53.6, 53.6, 53.6])
        metrics.log_confusion_matrix(['dog', 'cat', 'horses'],
                                     [[2, 6, 0], [3, 5, 6], [5, 7, 8]])
        with open(
                os.path.join(
                    os.path.dirname(__file__), 'test_data',
                    'expected_io_types_bulk_load_classification_metrics.json')
        ) as json_file:
            expected_json = json.load(json_file)
            self.assertEqual(expected_json, metrics.metadata)


@contextlib.contextmanager
def set_temporary_task_root(task_root: str):
    artifact_types.CONTAINER_TASK_ROOT = task_root
    try:
        yield
    finally:
        artifact_types.CONTAINER_TASK_ROOT = None


class TestGetUri(unittest.TestCase):

    def test_raise_if_no_env_var(self):

        with self.assertRaisesRegex(
                RuntimeError,
                r"'dsl\.get_uri' can only be called at task runtime\. The task root is unknown in the current environment\."
        ):
            dsl.get_uri()

    def test_default_gcs(self):
        with set_temporary_task_root(
                '/gcs/my_bucket/123456789/abc-09-14-2023-14-21-53/foo_123456789'
        ):
            self.assertEqual(
                'gs://my_bucket/123456789/abc-09-14-2023-14-21-53/foo_123456789/Output',
                dsl.get_uri())

    def test_default_s3(self):
        with set_temporary_task_root(
                '/s3/my_bucket/123456789/abc-09-14-2023-14-21-53/foo_123456789'
        ):
            self.assertEqual(
                's3://my_bucket/123456789/abc-09-14-2023-14-21-53/foo_123456789/Output',
                dsl.get_uri())

    def test_default_minio(self):
        with set_temporary_task_root(
                '/minio/my_bucket/123456789/abc-09-14-2023-14-21-53/foo_123456789'
        ):
            self.assertEqual(
                'minio://my_bucket/123456789/abc-09-14-2023-14-21-53/foo_123456789/Output',
                dsl.get_uri())

    def test_suffix_arg_gcs(self):
        with set_temporary_task_root(
                '/gcs/my_bucket/123456789/abc-09-14-2023-14-21-53/foo_123456789'
        ):
            self.assertEqual(
                'gs://my_bucket/123456789/abc-09-14-2023-14-21-53/foo_123456789/model',
                dsl.get_uri('model'))

    def test_suffix_arg_tmp_no_suffix(self):
        with set_temporary_task_root('/tmp/kfp_outputs'):
            with self.assertWarnsRegex(
                    RuntimeWarning,
                    r'dsl\.get_uri is not yet supported by the KFP backend\. Please specify a URI explicitly\.'
            ):
                actual = dsl.get_uri('model')
                self.assertEqual('', actual)

    def test_suffix_arg_tmp_with_suffix(self):
        with set_temporary_task_root('/tmp/kfp_outputs'):
            with self.assertWarnsRegex(
                    RuntimeWarning,
                    r'dsl\.get_uri is not yet supported by the KFP backend\. Please specify a URI explicitly\.'
            ):
                actual = dsl.get_uri('model')
                self.assertEqual('', actual)


class TestConvertLocalPathToRemotePath(parameterized.TestCase):

    @parameterized.parameters([{
        'local_path': local_path,
        'expected': expected
    } for local_path, expected in [
        ('/gcs/foo/bar', 'gs://foo/bar'),
        ('/minio/foo/bar', 'minio://foo/bar'),
        ('/s3/foo/bar', 's3://foo/bar'),
        ('/oci/quay.io_org_repo:latest/models',
         'oci://quay.io/org/repo:latest'),
        ('/oci/quay.io_org_repo:latest', 'oci://quay.io/org/repo:latest'),
        ('/tmp/kfp_outputs', '/tmp/kfp_outputs'),
        ('/some/random/path', '/some/random/path'),
    ]])
    def test_gcs(self, local_path, expected):
        actual = artifact_types.convert_local_path_to_remote_path(local_path)
        self.assertEqual(actual, expected)


if __name__ == '__main__':
    unittest.main()



# Copyright 2023 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for placeholder_utils.py."""

import json
from typing import List, Optional
import unittest

from absl.testing import parameterized
from google.protobuf import json_format
from kfp.local import placeholder_utils
from kfp.pipeline_spec import pipeline_spec_pb2

executor_input = pipeline_spec_pb2.ExecutorInput()
json_format.ParseDict(
    {
        'inputs': {
            'parameterValues': {
                'boolean': False,
                'dictionary': {
                    'foo': 'bar'
                },
            },
            'artifacts': {
                'in_a': {
                    'artifacts': [{
                        'name':
                            'in_a',
                        'type': {
                            'schemaTitle': 'system.Dataset',
                            'schemaVersion': '0.0.1'
                        },
                        'uri':
                            '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/upstream-comp/in_a',
                        'metadata': {
                            'foo': {
                                'bar': 'baz'
                            }
                        }
                    }]
                }
            },
        },
        'outputs': {
            'parameters': {
                'Output': {
                    'outputFile':
                        '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/comp/Output'
                }
            },
            'artifacts': {
                'out_a': {
                    'artifacts': [{
                        'name':
                            'out_a',
                        'type': {
                            'schemaTitle': 'system.Dataset',
                            'schemaVersion': '0.0.1'
                        },
                        'uri':
                            '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/comp/out_a',
                        # include metadata on outputs since it allows us to
                        # test the placeholder
                        # "{{$.outputs.artifacts[''out_a''].metadata[''foo'']}}"
                        # for comprehensive testing, but in practice metadata
                        # will never be set on output artifacts since they
                        # haven't been created yet
                        'metadata': {
                            'foo': {
                                'bar': 'baz'
                            }
                        }
                    }]
                }
            },
            'outputFile':
                '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/comp/executor_output.json'
        }
    },
    executor_input)

EXECUTOR_INPUT_DICT = json_format.MessageToDict(executor_input)


class TestReplacePlaceholders(unittest.TestCase):
    # most of the logic is tested in TestReplacePlaceholderForElement, so this is just a basic test to invoke the code and make sure the placeholder resolution is applied correctly to every element in the list

    def test(self):
        full_command = [
            'echo',
            'something before the placeholder {{$}}',
            'something else',
            '{{$.outputs.output_file}}',
        ]
        actual = placeholder_utils.replace_placeholders(
            full_command=full_command,
            executor_input_dict=EXECUTOR_INPUT_DICT,
            pipeline_resource_name='my-pipeline-2023-10-10-13-32-59-420710',
            task_resource_name='comp',
            pipeline_root='/foo/bar/my-pipeline-2023-10-10-13-32-59-420710',
            unique_pipeline_id=placeholder_utils.make_random_id(),
        )
        expected = [
            'echo',
            f'something before the placeholder {json.dumps(EXECUTOR_INPUT_DICT)}',
            'something else',
            '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/comp/executor_output.json',
        ]
        self.assertEqual(actual, expected)


class TestResolveIndividualPlaceholder(parameterized.TestCase):

    # TODO: consider supporting JSON escape
    # TODO: update when output lists of artifacts are supported
    @parameterized.parameters([
        (
            '{{$}}',
            json.dumps(EXECUTOR_INPUT_DICT),
        ),
        (
            '{{$.outputs.output_file}}',
            '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/comp/executor_output.json',
        ),
        (
            '{{$.outputMetadataUri}}',
            '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/comp/executor_output.json',
        ),
        (
            '{{$.pipeline_job_name}}',
            'my-pipeline-2023-10-10-13-32-59-420710',
        ),
        (
            '{{$.pipeline_job_uuid}}',
            '123456789',
        ),
        (
            '{{$.pipeline_task_name}}',
            'comp',
        ),
        (
            '{{$.pipeline_task_uuid}}',
            '987654321',
        ),
        (
            '{{$.pipeline_root}}',
            '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710',
        ),
    ])
    def test_constant_placeholders(self, element: str, expected: str):
        actual = placeholder_utils.resolve_individual_placeholder(
            element=element,
            executor_input_dict=EXECUTOR_INPUT_DICT,
            pipeline_resource_name='my-pipeline-2023-10-10-13-32-59-420710',
            task_resource_name='comp',
            pipeline_root='/foo/bar/my-pipeline-2023-10-10-13-32-59-420710',
            pipeline_job_id='123456789',
            pipeline_task_id='987654321',
        )
        self.assertEqual(actual, expected)

    @parameterized.parameters([
        (
            '{{$}}invalidjson',
            json.dumps(EXECUTOR_INPUT_DICT) + 'invalidjson',
        ),
        (
            '{{$.pipeline_job_name}}/{{$.pipeline_task_name}}',
            'my-pipeline-2023-10-10-13-32-59-420710/comp',
        ),
        (
            '{{$.pipeline_root}}/foo/bar',
            '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/foo/bar',
        ),
    ])
    def test_concatenated_placeholders_resolve(self, element: str,
                                               expected: str):
        actual = placeholder_utils.resolve_individual_placeholder(
            element=element,
            executor_input_dict=EXECUTOR_INPUT_DICT,
            pipeline_resource_name='my-pipeline-2023-10-10-13-32-59-420710',
            task_resource_name='comp',
            pipeline_root='/foo/bar/my-pipeline-2023-10-10-13-32-59-420710',
            pipeline_job_id='123456789',
            pipeline_task_id='987654321',
        )
        self.assertEqual(actual, expected)

    @parameterized.parameters([
        (
            "{{$.inputs.parameters[''boolean'']}}",
            json.dumps(False),
        ),
        (
            "{{$.inputs.parameters[''not_present'']}}",
            json.dumps(None),
        ),
        (
            "{{$.outputs.artifacts[''out_a''].metadata}}",
            json.dumps({'foo': {
                'bar': 'baz'
            }}),
        ),
        (
            "{{$.outputs.parameters[''Output''].output_file}}",
            '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/comp/Output',
        ),
        (
            "{{$.outputs.artifacts[''out_a''].uri}}",
            '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/comp/out_a',
        ),
        (
            "{{$.outputs.artifacts[''out_a''].path}}",
            '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/comp/out_a',
        ),
        (
            "{{$.outputs.artifacts[''out_a''].metadata[''foo'']}}",
            json.dumps({'bar': 'baz'}),
        ),
        (
            "{{$.inputs.artifacts[''in_a''].metadata}}",
            json.dumps({'foo': {
                'bar': 'baz'
            }}),
        ),
        (
            "{{$.inputs.artifacts[''in_a''].uri}}",
            '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/upstream-comp/in_a',
        ),
        (
            "{{$.inputs.artifacts[''in_a''].path}}",
            '/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/upstream-comp/in_a',
        ),
        (
            "{{$.inputs.artifacts[''in_a''].metadata[''foo'']}}",
            json.dumps({'bar': 'baz'}),
        ),
    ])
    def test_io_placeholders(self, element: str, expected: str):
        actual = placeholder_utils.resolve_individual_placeholder(
            element=element,
            executor_input_dict=EXECUTOR_INPUT_DICT,
            pipeline_resource_name='my-pipeline-2023-10-10-13-32-59-420710',
            task_resource_name='comp',
            pipeline_root='/foo/bar/my-pipeline-2023-10-10-13-32-59-420710',
            pipeline_job_id='123456789',
            pipeline_task_id='987654321',
        )
        self.assertEqual(actual, expected)

    @parameterized.parameters([
        (
            "my-prefix-{{$.inputs.parameters[''boolean'']}}-suffix",
            'my-prefix-false-suffix',
        ),
        (
            "--param={{$.inputs.parameters[''not_present'']}}",
            '--param=null',
        ),
        (
            "prefix{{$.outputs.parameters[''Output''].output_file}}/suffix",
            'prefix/foo/bar/my-pipeline-2023-10-10-13-32-59-420710/comp/Output/suffix',
        ),
        (
            "prefix{{$.inputs.parameters[''dictionary'']}}suffix",
            'prefix{"foo": "bar"}suffix',
        ),
    ])
    def test_io_placeholder_with_string_concat(self, element: str,
                                               expected: str):
        actual = placeholder_utils.resolve_individual_placeholder(
            element=element,
            executor_input_dict=EXECUTOR_INPUT_DICT,
            pipeline_resource_name='my-pipeline-2023-10-10-13-32-59-420710',
            task_resource_name='comp',
            pipeline_root='/foo/bar/my-pipeline-2023-10-10-13-32-59-420710',
            pipeline_job_id='123456789',
            pipeline_task_id='987654321',
        )
        self.assertEqual(actual, expected)


class TestGetValueUsingPath(unittest.TestCase):

    def test_valid_path(self):
        actual = placeholder_utils.get_value_using_path(
            {'a': {
                'b': {
                    'c': 10
                }
            }},
            ['a', 'b', 'c'],
        )
        expected = 10
        self.assertEqual(actual, expected)

    def test_invalid_path(self):
        actual = placeholder_utils.get_value_using_path(
            {'a': {
                'b': {
                    'c': 10
                }
            }},
            ['a', 'x'],
        )
        self.assertIsNone(actual)

    def test_empty_path(self):
        with self.assertRaisesRegex(ValueError, r'path cannot be empty\.'):
            placeholder_utils.get_value_using_path({'a': 20}, [])


class TestResolveStructPlaceholders(parameterized.TestCase):

    @parameterized.parameters([
        (
            """{"Concat": ["a", "b", "c"]}""",
            [],
            'abc',
        ),
        (
            """{"Concat": ["prefix", "-", "{{$.outputs.artifacts[''x''].uri}}"]}""",
            [],
            "prefix-{{$.outputs.artifacts[''x''].uri}}",
        ),
        (
            """{"Concat": ["a", {"Concat": ["b", "c"]}]}""",
            [],
            'abc',
        ),
        (
            """{"IfPresent": {"InputName": "x", "Then": ["foo"], "Else": ["bar"]}}""",
            [],
            ['bar'],
        ),
        (
            """{"IfPresent": {"InputName": "x", "Then": ["foo"], "Else": ["bar"]}}""",
            ['x'],
            ['foo'],
        ),
        (
            """{"Concat": ["a", {"Concat": ["b", {"Concat": ["c", "{{$.inputs.parameters[''input2'']}}"]}]}]}""",
            [],
            "abc{{$.inputs.parameters[''input2'']}}",
        ),
        (
            """{"Concat": ["a", {"Concat": ["b", {"IfPresent": {"InputName": "foo", "Then": "c", "Else": "d"}}]}]}""",
            [],
            'abd',
        ),
        (
            """{"Concat": ["--flag", {"Concat": ["=", {"IfPresent": {"InputName": "x", "Then": "thing", "Else": "otherwise"}}]}]}""",
            ['x'],
            '--flag=thing',
        ),
        (
            """{"Concat": ["a", {"IfPresent": {"InputName": "foo", "Then": {"Concat": ["--", "flag", "{{$.inputs.artifacts['input2'].path}}"]}, "Else": "b"}}, "c"]}""",
            [],
            'abc',
        ),
        (
            """{"Concat": ["--flag", {"IfPresent": {"InputName": "foo", "Then": {"Concat": ["=", "{{$.inputs.artifacts['input2'].path}}"]}, "Else": "b"}}, "-suffix"]}""",
            ['foo'],
            "--flag={{$.inputs.artifacts['input2'].path}}-suffix",
        ),
        (
            """{"Concat": ["a-", {"IfPresent": {"InputName": "foo", "Then": {"Concat": ["--", "flag"]}, "Else": "{{$.inputs.artifacts['input2'].path}}"}}, "-c"]}""",
            [],
            "a-{{$.inputs.artifacts['input2'].path}}-c",
        ),
        (
            """{"Concat": ["--", {"IfPresent": {"InputName": "foo", "Then": {"Concat": ["flag"]}, "Else": "{{$.inputs.artifacts['input2'].path}}"}}, "=c"]}""",
            ['foo'],
            '--flag=c',
        ),
        (
            """{"Concat": ["--", {"IfPresent": {"InputName": "foo", "Then": {"Concat": ["flag"]}}}, "=c"]}""",
            ['foo'],
            '--flag=c',
        ),
        (
            """{"Concat": ["--flag", {"IfPresent": {"InputName": "foo", "Then": {"Concat": ["=", "other", "_val"]}}}, "=foo"]}""",
            [],
            '--flag=foo',
        ),
        (
            """{"IfPresent": {"InputName": "foo", "Then": {"Concat": ["--", "flag"]}}}""",
            ['foo'],
            '--flag',
        ),
        (
            """{"IfPresent": {"InputName": "foo", "Then": {"Concat": ["--", "flag"]}}}""",
            [],
            None,
        ),
    ])
    def test(
        self,
        placeholder: str,
        provided_inputs: List[str],
        expected: Optional[None],
    ):
        actual = placeholder_utils.resolve_struct_placeholders(
            placeholder,
            provided_inputs,
        )
        self.assertEqual(actual, expected)


class TestResolveSelfReferencesInExecutorInput(unittest.TestCase):

    def test_simple(self):
        executor_input_dict = {
            'inputs': {
                'parameterValues': {
                    'pipelinechannel--identity-Output':
                        'foo',
                    'string':
                        "{{$.inputs.parameters['pipelinechannel--identity-Output']}}-bar"
                }
            },
            'outputs': {
                'outputFile':
                    '/foo/bar/my-pipeline-2024-01-26-12-26-24-162075/echo/executor_output.json'
            }
        }
        expected = {
            'inputs': {
                'parameterValues': {
                    'pipelinechannel--identity-Output': 'foo',
                    'string': 'foo-bar'
                }
            },
            'outputs': {
                'outputFile':
                    '/foo/bar/my-pipeline-2024-01-26-12-26-24-162075/echo/executor_output.json'
            }
        }
        actual = placeholder_utils.resolve_self_references_in_executor_input(
            executor_input_dict,
            pipeline_resource_name='my-pipeline-2024-01-26-12-26-24-162075',
            task_resource_name='echo',
            pipeline_root='/foo/bar/my-pipeline-2024-01-26-12-26-24-162075',
            pipeline_job_id='123456789',
            pipeline_task_id='987654321',
        )
        self.assertEqual(actual, expected)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes for input/output type annotations in KFP SDK.

These are only compatible with v2 Pipelines.
"""

import re
from typing import Any, List, Optional, Type, TypeVar, Union

from kfp.dsl.types import artifact_types
from kfp.dsl.types import type_annotations
from kfp.dsl.types import type_utils


class OutputPath:
    """Type annotation used in component definitions for indicating a parameter
    is a path to an output. The path parameter typed with this annotation can
    be treated as a locally accessible filepath within the component body.

    The argument typed with this annotation is provided at runtime by the executing backend and does not need to be passed as an input by the pipeline author (see example).


    Args:
        type: The type of the value written to the output path.

    Example:
      ::

        @dsl.component
        def create_parameter(
                message: str,
                output_parameter_path: OutputPath(str),
        ):
            with open(output_parameter_path, 'w') as f:
                f.write(message)


        @dsl.component
        def consume_parameter(message: str):
            print(message)


        @dsl.pipeline(name='my-pipeline', pipeline_root='gs://my-bucket')
        def my_pipeline(message: str = 'default message'):
            create_param_op = create_parameter(message=message)
            consume_parameter(message=create_param_op.outputs['output_parameter_path'])
    """

    def __init__(self, type=None):
        self.type = construct_type_for_inputpath_or_outputpath(type)

    def __eq__(self, other):
        return isinstance(other, OutputPath) and self.type == other.type


class InputPath:
    """Type annotation used in component definitions for indicating a parameter
    is a path to an input.

    Example:
      ::

        @dsl.component
        def create_dataset(dataset_path: OutputPath('Dataset'),):
            import json
            dataset = {'my_dataset': [[1, 2, 3], [4, 5, 6]]}
            with open(dataset_path, 'w') as f:
                json.dump(dataset, f)


        @dsl.component
        def consume_dataset(dataset: InputPath('Dataset')):
            print(dataset)


        @dsl.pipeline(name='my-pipeline', pipeline_root='gs://my-bucket')
        def my_pipeline():
            create_dataset_op = create_dataset()
            consume_dataset(dataset=create_dataset_op.outputs['dataset_path'])
    """

    def __init__(self, type=None):
        self.type = construct_type_for_inputpath_or_outputpath(type)

    def __eq__(self, other):
        return isinstance(other, InputPath) and self.type == other.type


def construct_type_for_inputpath_or_outputpath(
        type_: Union[str, Type, None]) -> Optional[str]:
    if type_annotations.is_artifact_class(type_):
        return type_utils.create_bundled_artifact_type(type_.schema_title,
                                                       type_.schema_version)
    elif isinstance(
            type_,
            str) and type_.lower() in type_utils.ARTIFACT_CLASSES_MAPPING:
        # v1 artifact backward compat, e.g. dsl.OutputPath('Dataset')
        return type_utils.create_bundled_artifact_type(
            type_utils.ARTIFACT_CLASSES_MAPPING[type_.lower()].schema_title)
    elif type_utils.get_parameter_type(type_):
        return type_
    else:
        # v1 unknown type dsl.OutputPath('MyCustomType')
        return type_utils.create_bundled_artifact_type(
            artifact_types.Artifact.schema_title)


class InputAnnotation:
    """Marker type for input artifacts."""


class OutputAnnotation:
    """Marker type for output artifacts."""


def is_Input_Output_artifact_annotation(typ) -> bool:
    if not hasattr(typ, '__metadata__'):
        return False

    if typ.__metadata__[0] not in [InputAnnotation, OutputAnnotation]:
        return False

    return True


def is_artifact_wrapped_in_Input(typ: Any) -> bool:
    """Returns True if typ is of type Input[T]."""
    if not is_Input_Output_artifact_annotation(typ):
        return False

    return typ.__metadata__[0] == InputAnnotation


def is_artifact_wrapped_in_Output(typ: Any) -> bool:
    """Returns True if typ is of type Output[T]."""
    if not is_Input_Output_artifact_annotation(typ):
        return False

    return typ.__metadata__[0] == OutputAnnotation


def get_io_artifact_class(typ):
    from kfp.dsl import Input
    from kfp.dsl import Output
    if not is_Input_Output_artifact_annotation(typ):
        return None
    if typ == Input or typ == Output:
        return None

    # extract inner type from list of artifacts
    inner = strip_Input_or_Output_marker(typ)
    if hasattr(inner, '__origin__') and inner.__origin__ == list:
        return inner.__args__[0]

    return inner


def strip_Input_or_Output_marker(typ: Any) -> artifact_types.Artifact:
    return typ.__args__[0]


def get_input_or_output_marker(
        typ) -> Optional[Union[InputAnnotation, OutputAnnotation]]:
    if not is_Input_Output_artifact_annotation(typ):
        return None

    return typ.__metadata__[0]


T = TypeVar('T')


def maybe_strip_optional_from_annotation(annotation: T) -> T:
    """Strips 'Optional' from 'Optional[<type>]' if applicable.

    For example::
      Optional[str] -> str
      str -> str
      List[int] -> List[int]

    Args:
      annotation: The original type annotation which may or may not has
        `Optional`.

    Returns:
      The type inside Optional[] if Optional exists, otherwise the original type.
    """
    if getattr(annotation, '__origin__',
               None) is Union and annotation.__args__[1] is type(None):
        return annotation.__args__[0]
    return annotation


def maybe_strip_optional_from_annotation_string(annotation: str) -> str:
    if annotation.startswith('Optional[') and annotation.endswith(']'):
        return annotation.lstrip('Optional[').rstrip(']')
    return annotation


def get_short_type_name(type_name: str) -> str:
    """Extracts the short form type name.

    This method is used for looking up serializer for a given type.

    For example::
      typing.List -> List
      typing.List[int] -> List
      typing.Dict[str, str] -> Dict
      List -> List
      str -> str

    Args:
      type_name: The original type name.

    Returns:
      The short form type name or the original name if pattern doesn't match.
    """
    match = re.match(r'(typing\.)?(?P<type>\w+)(?:\[.+\])?', type_name)
    return match['type'] if match else type_name


def is_artifact_class(artifact_class_or_instance: Type) -> bool:
    # we do not yet support non-pre-registered custom artifact types with instance_schema attribute
    return hasattr(artifact_class_or_instance, 'schema_title') and hasattr(
        artifact_class_or_instance, 'schema_version')


def is_list_of_artifacts(
    annotation: Union[Type[List[artifact_types.Artifact]],
                      Type[artifact_types.Artifact]]
) -> bool:
    """Checks if an object is a list of list of artifacts annotation (e.g.,
    List[Artifact], List[Dataset])"""
    return is_generic_list(annotation) and issubclass_of_artifact(
        get_inner_type(annotation))


def get_inner_type(annotation: Any) -> Optional[Any]:
    """Returns the inner type of a generic annotation.

    For Union or Optional types with multiple inner types, a tuple of
    types is returned.
    """
    # Check if the annotation has '__args__' attribute
    if hasattr(annotation, '__args__'):
        if len(annotation.__args__) == 1:
            return annotation.__args__[0]
        else:
            return tuple(annotation.__args__)
    return None


def issubclass_of_artifact(obj: Any) -> bool:
    """Checks if an object is a class and a subclass of a dsl.Artifact."""
    return type(obj) == type and issubclass(obj, artifact_types.Artifact)


def is_generic_list(annotation: Any) -> bool:
    # handles generics from the typing module for python<3.9
    typing_generic_list = getattr(annotation, '__origin__',
                                  None) is list or getattr(
                                      annotation, '__origin__', None) is List
    # handles built-in generics for python>=3.9
    built_in_generic_list = annotation == list
    return typing_generic_list or built_in_generic_list



# Copyright 2020-2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for component I/O type mapping."""

import inspect
import json
from typing import Any, Callable, Dict, Optional, Type, Union
import warnings

import kfp
from kfp.dsl import task_final_status
from kfp.dsl.types import artifact_types
from kfp.dsl.types import type_annotations

DEFAULT_ARTIFACT_SCHEMA_VERSION = '0.0.1'
PARAMETER_TYPES = Union[str, int, float, bool, dict, list]

# ComponentSpec I/O types to DSL ontology artifact classes mapping.
ARTIFACT_CLASSES_MAPPING = {
    'artifact': artifact_types.Artifact,
    'model': artifact_types.Model,
    'dataset': artifact_types.Dataset,
    'metrics': artifact_types.Metrics,
    'classificationmetrics': artifact_types.ClassificationMetrics,
    'slicedclassificationmetrics': artifact_types.SlicedClassificationMetrics,
    'html': artifact_types.HTML,
    'markdown': artifact_types.Markdown,
}

_GOOGLE_TYPES_PATTERN = r'^google.[A-Za-z]+$'
_GOOGLE_TYPES_VERSION = DEFAULT_ARTIFACT_SCHEMA_VERSION

# ComponentSpec I/O types to (IR) PipelineTaskSpec I/O types mapping.
# The keys are normalized (lowercased). These are types viewed as Parameters.
# The values are the corresponding IR parameter primitive types.

# pipeline_spec_pb2.ParameterType enum values
NUMBER_DOUBLE = 1
NUMBER_INTEGER = 2
STRING = 3
BOOLEAN = 4
LIST = 5
STRUCT = 6
TASK_FINAL_STATUS = 7
PARAMETER_TYPES_MAPPING = {
    'integer': NUMBER_INTEGER,
    'int': NUMBER_INTEGER,
    'double': NUMBER_DOUBLE,
    'float': NUMBER_DOUBLE,
    'string': STRING,
    'str': STRING,
    'text': STRING,
    'bool': BOOLEAN,
    'boolean': BOOLEAN,
    'dict': STRUCT,
    'list': LIST,
    'jsonobject': STRUCT,
    'jsonarray': LIST,
}


# copied from distutils.util, which was removed in Python 3.12
# https://github.com/pypa/distutils/blob/fb5c5704962cd3f40c69955437da9a88f4b28567/distutils/util.py#L340-L353
def strtobool(val):
    """Convert a string representation of truth to true (1) or false (0).

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.
    """
    val = val.lower()
    if val in ('y', 'yes', 't', 'true', 'on', '1'):
        return 1
    elif val in ('n', 'no', 'f', 'false', 'off', '0'):
        return 0
    else:
        raise ValueError('invalid truth value %r' % (val,))


def bool_cast_fn(default: Union[str, bool]) -> bool:
    if isinstance(default, str):
        default = strtobool(default) == 1
    return default


def try_loading_json(default: str) -> Union[dict, list, str]:
    try:
        return json.loads(default)
    except:
        return default


_V1_DEFAULT_DESERIALIZER_MAPPING: Dict[str, Callable] = {
    'integer': int,
    'int': int,
    'double': float,
    'float': float,
    'string': str,
    'str': str,
    'text': str,
    'bool': bool_cast_fn,
    'boolean': bool_cast_fn,
    'dict': try_loading_json,
    'list': try_loading_json,
    'jsonobject': try_loading_json,
    'jsonarray': try_loading_json,
}


def deserialize_v1_component_yaml_default(type_: str, default: Any) -> Any:
    """Deserializes v1 default values to correct in-memory types.

    Typecasts for primitive types. Tries to load JSON for arrays and
    structs.
    """
    if default is None:
        return default
    if isinstance(type_, str):
        cast_fn = _V1_DEFAULT_DESERIALIZER_MAPPING.get(type_.lower(),
                                                       lambda x: x)
        return cast_fn(default)
    return default


def is_task_final_status_type(type_name: Optional[Union[str, dict]]) -> bool:
    """Check if a ComponentSpec I/O type is PipelineTaskFinalStatus.

    Args:
      type_name: type name of the ComponentSpec I/O type.

    Returns:
      True if the type name is 'PipelineTaskFinalStatus'.
    """
    return isinstance(type_name, str) and (
        type_name == task_final_status.PipelineTaskFinalStatus.__name__)


def is_parameter_type(type_name: Optional[Union[str, dict]]) -> bool:
    """Check if a ComponentSpec I/O type is considered as a parameter type.

    Args:
      type_name: type name of the ComponentSpec I/O type.

    Returns:
      True if the type name maps to a parameter type else False.
    """
    if isinstance(type_name, str):
        type_name = type_annotations.get_short_type_name(type_name)
    elif isinstance(type_name, dict):
        type_name = list(type_name.keys())[0]
    else:
        return False

    return type_name.lower(
    ) in PARAMETER_TYPES_MAPPING or is_task_final_status_type(type_name)


def bundled_artifact_to_artifact_proto(
        bundled_artifact_str: str) -> 'pipeline_spec_pb2.ArtifactTypeSchema':
    """Gets the IR ArtifactTypeSchema proto for a bundled artifact in form
    `<namespace>.<Name>@x.x.x` (e.g., system.Artifact@0.0.1)."""
    bundled_artifact_str, schema_version = bundled_artifact_str.split('@')

    from kfp.pipeline_spec import pipeline_spec_pb2

    return pipeline_spec_pb2.ArtifactTypeSchema(
        schema_title=bundled_artifact_str,
        schema_version=schema_version,
    )


def get_parameter_type(
    param_type: Optional[Union[Type, str, dict]]
) -> 'pipeline_spec_pb2.ParameterType':
    """Get the IR I/O parameter type for the given ComponentSpec I/O type.

    Args:
      param_type: type of the ComponentSpec I/O type. Can be a primitive Python
        builtin type or a type name.

    Returns:
      The enum value of the mapped IR I/O primitive type.

    Raises:
      AttributeError: if type_name is not a string type.
    """
    # Special handling for PipelineTaskFinalStatus, treat it as Dict type.
    if is_task_final_status_type(param_type):
        param_type = 'dict'
    if type(param_type) == type:
        type_name = param_type.__name__
    elif isinstance(param_type, dict):
        type_name = list(param_type.keys())[0]
    else:
        type_name = type_annotations.get_short_type_name(str(param_type))
    return PARAMETER_TYPES_MAPPING.get(type_name.lower())


def get_parameter_type_name(
        param_type: Optional[Union[Type, str, dict]]) -> str:
    """Gets the parameter type name."""

    from kfp.pipeline_spec import pipeline_spec_pb2
    param_enum_val = get_parameter_type(param_type)
    if param_enum_val is None:
        raise ValueError(
            '`param_type` is not a parameter type. Cannot get ParameterType name.'
        )
    return pipeline_spec_pb2.ParameterType.ParameterTypeEnum.Name(
        param_enum_val)


class InconsistentTypeException(Exception):
    """InconsistencyTypeException is raised when two types are not
    consistent."""


class InconsistentTypeWarning(Warning):
    """InconsistentTypeWarning is issued when two types are not consistent."""


def _get_type_string_from_component_argument(
    argument_value: Union['pipeline_channel.PipelineChannel', str, bool, int,
                          float, dict, list]
) -> str:
    # avoid circular imports
    from kfp.dsl import pipeline_channel
    if isinstance(argument_value, pipeline_channel.PipelineChannel):
        return argument_value.channel_type

    # argument is a constant
    argument_type = type(argument_value)
    if argument_type in _TYPE_TO_TYPE_NAME:
        return _TYPE_TO_TYPE_NAME[argument_type]

    if isinstance(argument_value, artifact_types.Artifact):
        raise ValueError(
            f'Input artifacts are not supported. Got input artifact of type {argument_value.__class__.__name__!r}.'
        )
    raise ValueError(
        f'Constant argument inputs must be one of type {list(_TYPE_TO_TYPE_NAME.values())} Got: {argument_value!r} of type {type(argument_value)!r}.'
    )


def verify_type_compatibility(
    given_value: Union['pipeline_channel.PipelineChannel', str, bool, int,
                       float, dict, list],
    expected_spec: Union['structures.InputSpec', 'structures.OutputSpec'],
    error_message_prefix: str,
    checks_input: bool = True,
    raise_on_error: bool = True,
) -> bool:
    """Verifies the given argument type is compatible with the expected type.

    Args:
        given_value: The channel or constant provided as an argument.
        expected_spec: The InputSpec or OutputSpec that describes the expected type of given_value.
        error_message_prefix: The prefix for the error message.
        checks_input: True if checks an argument (given_value) against a component/pipeline input type (expected_spec). False if checks a component output (argument_value) against the pipeline output type (expected_spec).
        raise_on_error: Whether to raise on type compatibility error. Should be passed kfp.TYPE_CHECK.

    Returns:
        True if types are compatible, and False if otherwise.

    Raises:
        InconsistentTypeException if raise_on_error=True.
    """
    # extract and normalize types
    expected_type = expected_spec.type
    given_type = _get_type_string_from_component_argument(given_value)

    # avoid circular imports
    from kfp.dsl import for_loop

    # Workaround for potential type-checking issues during ParallelFor compilation: When LoopArgument or LoopArgumentVariable are involved and the expected type is 'String', we temporarily relax type enforcement to avoid blocking compilation. This is necessary due to potential information loss during the compilation step.
    if isinstance(given_value,
                  (for_loop.LoopParameterArgument,
                   for_loop.LoopArgumentVariable)) and given_type == 'String':
        return True

    given_is_param = is_parameter_type(str(given_type))
    if given_is_param:
        given_type = get_parameter_type_name(given_type)
        given_is_artifact_list = False
    else:
        given_is_artifact_list = given_value.is_artifact_list

    expected_is_param = is_parameter_type(expected_type)
    if expected_is_param:
        expected_type = get_parameter_type_name(expected_type)
        expected_is_artifact_list = False
    else:
        expected_is_artifact_list = expected_spec.is_artifact_list

    # compare the normalized types
    if given_is_param != expected_is_param:
        types_are_compatible = False
    elif given_is_param and expected_is_param:
        types_are_compatible = check_parameter_type_compatibility(
            given_type, expected_type)
    else:
        types_are_compatible = check_artifact_type_compatibility(
            given_type=given_type,
            given_is_artifact_list=given_is_artifact_list,
            expected_type=expected_type,
            expected_is_artifact_list=expected_is_artifact_list)

    # maybe raise, maybe warn, return bool
    if not types_are_compatible:
        # update the types for lists of artifacts for error message
        given_type = f'List[{given_type}]' if given_is_artifact_list else given_type
        expected_type = f'List[{expected_type}]' if expected_is_artifact_list else expected_type
        if checks_input:
            error_message_suffix = f'Argument type {given_type!r} is incompatible with the input type {expected_type!r}'
        else:
            error_message_suffix = f'Output of type {given_type!r} cannot be surfaced as pipeline output type {expected_type!r}'
        error_text = error_message_prefix + error_message_suffix
        if raise_on_error:
            raise InconsistentTypeException(error_text)
        else:
            warnings.warn(InconsistentTypeWarning(error_text))

    return types_are_compatible


def check_artifact_type_compatibility(given_type: str,
                                      given_is_artifact_list: bool,
                                      expected_type: str,
                                      expected_is_artifact_list: bool) -> bool:
    given_schema_title, given_schema_version = given_type.split('@')
    expected_schema_title, expected_schema_version = expected_type.split('@')
    same_list_of_artifacts_status = expected_is_artifact_list == given_is_artifact_list
    if not same_list_of_artifacts_status:
        return False
    elif artifact_types.Artifact.schema_title in {
            given_schema_title, expected_schema_title
    }:
        return True
    else:
        schema_title_compatible = given_schema_title == expected_schema_title
        schema_version_compatible = given_schema_version.split(
            '.')[0] == expected_schema_version.split('.')[0]

        return schema_title_compatible and schema_version_compatible


def check_parameter_type_compatibility(given_type: str,
                                       expected_type: str) -> bool:
    if isinstance(given_type, str) and isinstance(expected_type, str):
        return given_type == expected_type
    else:
        return check_v1_struct_parameter_type_compatibility(
            given_type, expected_type)


def check_v1_struct_parameter_type_compatibility(
    given_type: Union[str, dict],
    expected_type: Union[str, dict],
) -> bool:
    if isinstance(given_type, str):
        given_type = {given_type: {}}
    if isinstance(expected_type, str):
        expected_type = {expected_type: {}}
    return _check_dict_types(given_type, expected_type)


def _check_dict_types(
    given_type: dict,
    expected_type: dict,
) -> bool:
    given_type_name, _ = list(given_type.items())[0]
    expected_type_name, _ = list(expected_type.items())[0]
    if given_type_name == '' or expected_type_name == '':
        # If the type name is empty, it matches any types
        return True
    if given_type_name != expected_type_name:
        print('type name ' + str(given_type_name) +
              ' is different from expected: ' + str(expected_type_name))
        return False
    type_name = given_type_name
    for type_property in given_type[type_name]:
        if type_property not in expected_type[type_name]:
            print(type_name + ' has a property ' + str(type_property) +
                  ' that the latter does not.')
            return False
        if given_type[type_name][type_property] != expected_type[type_name][
                type_property]:
            print(type_name + ' has a property ' + str(type_property) +
                  ' with value: ' + str(given_type[type_name][type_property]) +
                  ' and ' + str(expected_type[type_name][type_property]))
            return False
    return True


_TYPE_TO_TYPE_NAME = {
    str: 'String',
    int: 'Integer',
    float: 'Float',
    bool: 'Boolean',
    list: 'List',
    dict: 'Dict',
}


def get_canonical_type_name_for_type(typ: Type) -> Optional[str]:
    """Find the canonical type name for a given type.

    Args:
        typ: The type to search for.

    Returns:
        The canonical name of the type found.
    """
    return _TYPE_TO_TYPE_NAME.get(typ, None)


class TypeCheckManager:
    """Context manager to set a type check mode within context, then restore
    mode to original value upon exiting the context."""

    def __init__(self, enable: bool) -> None:
        """TypeCheckManager constructor.

        Args:
            enable: Type check mode used within context.
        """
        self._enable = enable

    def __enter__(self) -> 'TypeCheckManager':
        """Set type check mode to self._enable.

        Returns:
            TypeCheckManager: Returns itself.
        """
        self._prev = kfp.TYPE_CHECK
        kfp.TYPE_CHECK = self._enable
        return self

    def __exit__(self, *unused_args) -> None:
        """Restore type check mode to its previous state."""
        kfp.TYPE_CHECK = self._prev


# for reading in IR back to in-memory data structures
IR_TYPE_TO_IN_MEMORY_SPEC_TYPE = {
    'STRING': 'String',
    'NUMBER_INTEGER': 'Integer',
    'NUMBER_DOUBLE': 'Float',
    'LIST': 'List',
    'STRUCT': 'Dict',
    'BOOLEAN': 'Boolean',
    'TASK_FINAL_STATUS': task_final_status.PipelineTaskFinalStatus.__name__,
}

IR_TYPE_TO_COMMENT_TYPE_STRING = {
    'STRING': str.__name__,
    'NUMBER_INTEGER': int.__name__,
    'NUMBER_DOUBLE': float.__name__,
    'LIST': list.__name__,
    'STRUCT': dict.__name__,
    'BOOLEAN': bool.__name__,
    'TASK_FINAL_STATUS': task_final_status.PipelineTaskFinalStatus.__name__,
}

IN_MEMORY_SPEC_TYPE_TO_IR_TYPE = {
    v: k for k, v in IR_TYPE_TO_IN_MEMORY_SPEC_TYPE.items()
}


def get_canonical_name_for_outer_generic(type_name: Any) -> str:
    """Maps a complex/nested type name back to a canonical type.

    E.g.
        get_canonical_name_for_outer_generic('typing.List[str]')
        'List'

        get_canonical_name_for_outer_generic('typing.Dict[typing.List[str], str]')
        'Dict'

    Args:
        type_name (Any): The type. Returns input if not a string.

    Returns:
        str: The canonical type.
    """
    if not isinstance(type_name, str):
        return type_name

    if type_name.startswith('typing.'):
        type_name = type_name.lstrip('typing.')

    if type_name.lower().startswith('list') or type_name.lower().startswith(
            'dict'):
        return type_name.split('[')[0]

    else:
        return type_name


def create_bundled_artifact_type(schema_title: str,
                                 schema_version: Optional[str] = None) -> str:
    if not isinstance(schema_title, str):
        raise ValueError
    return schema_title + '@' + (
        schema_version or DEFAULT_ARTIFACT_SCHEMA_VERSION)


def validate_schema_version(schema_version: str) -> None:
    split_schema_version = schema_version.split('.')
    if len(split_schema_version) != 3:
        raise TypeError(
            f'Artifact schema_version must use three-part semantic versioning. Got: {schema_version}'
        )


def validate_schema_title(schema_title: str) -> None:
    split_schema_title = schema_title.split('.')
    if len(split_schema_title) != 2:
        raise TypeError(
            f'Artifact schema_title must have both a namespace and a name, separated by a `.`. Got: {schema_title}'
        )
    namespace, _ = split_schema_title
    if namespace not in {'system', 'google'}:
        raise TypeError(
            f'Artifact schema_title must belong to `system` or `google` namespace. Got: {schema_title}'
        )


def validate_bundled_artifact_type(type_: str) -> None:
    split_type = type_.split('@')
    # two parts and neither are empty strings
    if len(split_type) != 2 or not all(split_type):
        raise TypeError(
            f'Artifacts must have both a schema_title and a schema_version, separated by `@`. Got: {type_}'
        )
    schema_title, schema_version = split_type
    validate_schema_title(schema_title)
    validate_schema_version(schema_version)


def _annotation_to_type_struct(annotation):
    if not annotation or annotation == inspect.Parameter.empty:
        return None
    if hasattr(annotation, 'to_dict'):
        annotation = annotation.to_dict()
    if isinstance(annotation, dict):
        return annotation
    if isinstance(annotation, type):
        type_struct = get_canonical_type_name_for_type(annotation)
        if type_struct:
            return type_struct
        elif type_annotations.is_artifact_class(annotation):
            schema_title = annotation.schema_title
        else:
            schema_title = str(annotation.__name__)
    elif hasattr(annotation, '__forward_arg__'):
        schema_title = str(annotation.__forward_arg__)
    else:
        schema_title = str(annotation)
    type_struct = get_canonical_type_name_for_type(schema_title)
    return type_struct or schema_title


def is_typed_named_tuple_annotation(annotation: Any) -> bool:
    return hasattr(annotation, '_fields') and hasattr(annotation,
                                                      '__annotations__')



# Copyright 2023 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for logging_utils.py."""

import io
import unittest
from unittest import mock

from kfp import dsl
from kfp.local import logging_utils
from kfp.local import status


class TestIndentedPrint(unittest.TestCase):

    @mock.patch('sys.stdout', new_callable=io.StringIO)
    def test(self, mocked_stdout):
        with logging_utils.indented_print(num_spaces=6):
            print('foo should be indented')
        expected = '      foo should be indented\n'
        actual = mocked_stdout.getvalue()
        self.assertEqual(
            actual,
            expected,
        )


class TestColorText(unittest.TestCase):

    def test_cyan(self):

        actual = logging_utils.color_text(
            'text to color',
            logging_utils.Color.CYAN,
        )
        expected = '\x1b[91mtext to color\x1b[0m'
        self.assertEqual(actual, expected)

    def test_cyan(self):

        actual = logging_utils.color_text(
            'text to color',
            logging_utils.Color.RED,
        )
        expected = '\x1b[91mtext to color\x1b[0m'
        self.assertEqual(actual, expected)


class TestRenderArtifact(unittest.TestCase):

    def test_empty(self):
        actual = logging_utils.make_log_lines_for_artifact(dsl.Artifact())
        expected = [
            "Artifact( name='',",
            "          uri='',",
            '          metadata={} )',
        ]
        self.assertListEqual(actual, expected)

    def test_contains_value(self):
        actual = logging_utils.make_log_lines_for_artifact(
            dsl.Model(
                name='my_artifact',
                uri='/local/foo/bar',
                metadata={
                    'dict_field': {
                        'baz': 'bat'
                    },
                    'float_field': 3.14
                }))
        expected = [
            "Model( name='my_artifact',",
            "       uri='/local/foo/bar',",
            "       metadata={'dict_field': {'baz': 'bat'}, 'float_field': 3.14} )",
        ]
        self.assertListEqual(actual, expected)


class TestMakeLogLinesForOutputs(unittest.TestCase):

    def test_empty(self):
        actual = logging_utils.make_log_lines_for_outputs(dict())
        expected = []
        self.assertListEqual(actual, expected)

    def test_only_params(self):
        actual = logging_utils.make_log_lines_for_outputs({
            'foo': 'bar',
            'baz': 100,
            'bat': 1.0,
            'brap': True,
            'my_list': [1, 2, 3],
            'my_dict': {
                'foo': 'bar'
            }
        })
        expected = [
            "    foo: 'bar'",
            '    baz: 100',
            '    bat: 1.0',
            '    brap: True',
            '    my_list: [1, 2, 3]',
            "    my_dict: {'foo': 'bar'}",
        ]
        self.assertListEqual(actual, expected)

    def test_only_artifacts(self):
        actual = logging_utils.make_log_lines_for_outputs({
            'my_artifact':
                dsl.Artifact(name=''),
            'my_model':
                dsl.Model(
                    name='my_artifact',
                    uri='/local/foo/bar/1234567890/1234567890/1234567890/1234567890/1234567890',
                    metadata={
                        'dict_field': {
                            'baz': 'bat'
                        },
                        'float_field': 3.14
                    }),
            'my_dataset':
                dsl.Dataset(
                    name='my_dataset',
                    uri='/local/foo/baz',
                    metadata={},
                ),
        })
        expected = [
            "    my_artifact: Artifact( name='',",
            "                           uri='',",
            '                           metadata={} )',
            "    my_model: Model( name='my_artifact',",
            "                     uri='/local/foo/bar/1234567890/1234567890/1234567890/1234567890/1234567890',",
            "                     metadata={'dict_field': {'baz': 'bat'}, 'float_field': 3.14} )",
            "    my_dataset: Dataset( name='my_dataset',",
            "                         uri='/local/foo/baz',",
            '                         metadata={} )',
        ]
        self.assertListEqual(actual, expected)

    def test_mix_params_and_artifacts(self):
        actual = logging_utils.make_log_lines_for_outputs({
            'foo':
                'bar',
            'baz':
                100,
            'bat':
                1.0,
            'brap':
                True,
            'my_list': [1, 2, 3],
            'my_dict': {
                'foo': 'bar'
            },
            'my_artifact':
                dsl.Artifact(name=''),
            'my_model':
                dsl.Model(
                    name='my_artifact',
                    uri='/local/foo/bar/1234567890/1234567890/1234567890/1234567890/1234567890',
                    metadata={
                        'dict_field': {
                            'baz': 'bat'
                        },
                        'float_field': 3.14
                    }),
            'my_dataset':
                dsl.Dataset(
                    name='my_dataset',
                    uri='/local/foo/baz',
                    metadata={},
                ),
        })
        expected = [
            "    foo: 'bar'",
            '    baz: 100',
            '    bat: 1.0',
            '    brap: True',
            '    my_list: [1, 2, 3]',
            "    my_dict: {'foo': 'bar'}",
            "    my_artifact: Artifact( name='',",
            "                           uri='',",
            '                           metadata={} )',
            "    my_model: Model( name='my_artifact',",
            "                     uri='/local/foo/bar/1234567890/1234567890/1234567890/1234567890/1234567890',",
            "                     metadata={'dict_field': {'baz': 'bat'}, 'float_field': 3.14} )",
            "    my_dataset: Dataset( name='my_dataset',",
            "                         uri='/local/foo/baz',",
            '                         metadata={} )',
        ]

        self.assertListEqual(actual, expected)


class TestFormatStatus(unittest.TestCase):

    def test_success_status(self):
        self.assertEqual(
            logging_utils.format_status(status.Status.SUCCESS),
            '\x1b[92mSUCCESS\x1b[0m')

    def test_failure_status(self):
        self.assertEqual(
            logging_utils.format_status(status.Status.FAILURE),
            '\x1b[91mFAILURE\x1b[0m')

    def test_invalid_status(self):
        with self.assertRaisesRegex(ValueError,
                                    r'Got unknown status: INVALID_STATUS'):
            logging_utils.format_status('INVALID_STATUS')


class TestFormatTaskName(unittest.TestCase):

    def test(self):
        self.assertEqual(
            logging_utils.format_task_name('my-task'),
            '\x1b[96m\'my-task\'\x1b[0m')


class TestFormatPipelineName(unittest.TestCase):

    def test(self):
        self.assertEqual(
            logging_utils.format_pipeline_name('my-pipeline'),
            '\033[95m\'my-pipeline\'\033[0m')


if __name__ == '__main__':
    unittest.main()



# Copyright 2023 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Code for dispatching a local task execution."""
import logging
from typing import Any, Dict, Tuple

from kfp import local
from kfp.local import config
from kfp.local import docker_task_handler
from kfp.local import executor_input_utils
from kfp.local import executor_output_utils
from kfp.local import logging_utils
from kfp.local import placeholder_utils
from kfp.local import status
from kfp.local import subprocess_task_handler
from kfp.local import task_handler_interface
from kfp.local import utils
from kfp.pipeline_spec import pipeline_spec_pb2


def run_single_task(
    pipeline_spec: pipeline_spec_pb2.PipelineSpec,
    arguments: Dict[str, Any],
) -> Dict[str, Any]:
    """Runs a single component from its compiled PipelineSpec.

    Args:
        pipeline_spec: The PipelineSpec of the component to run.
        arguments: The runtime arguments.

    Returns:
        A LocalTask instance.
    """
    config.LocalExecutionConfig.validate()
    component_name, component_spec = list(pipeline_spec.components.items())[0]
    executor_spec = get_executor_spec(
        pipeline_spec,
        component_spec.executor_label,
    )
    executor_spec = utils.struct_to_executor_spec(executor_spec)
    pipeline_resource_name = executor_input_utils.get_local_pipeline_resource_name(
        pipeline_spec.pipeline_info.name)

    # all global state should be accessed here
    # do not access local config state downstream
    outputs, _ = run_single_task_implementation(
        pipeline_resource_name=pipeline_resource_name,
        component_name=component_name,
        component_spec=component_spec,
        executor_spec=executor_spec,
        arguments=arguments,
        pipeline_root=config.LocalExecutionConfig.instance.pipeline_root,
        runner=config.LocalExecutionConfig.instance.runner,
        raise_on_error=config.LocalExecutionConfig.instance.raise_on_error,
        block_input_artifact=True,
        unique_pipeline_id=placeholder_utils.make_random_id())
    return outputs


def get_executor_spec(
    pipeline_spec: pipeline_spec_pb2.PipelineSpec,
    executor_label: str,
) -> pipeline_spec_pb2.PipelineDeploymentConfig.ExecutorSpec:
    return pipeline_spec.deployment_spec['executors'][executor_label]


Outputs = Dict[str, Any]


def run_single_task_implementation(
    pipeline_resource_name: str,
    component_name: str,
    component_spec: pipeline_spec_pb2.ComponentSpec,
    executor_spec: pipeline_spec_pb2.PipelineDeploymentConfig.ExecutorSpec,
    arguments: Dict[str, Any],
    pipeline_root: str,
    runner: config.LocalRunnerType,
    raise_on_error: bool,
    block_input_artifact: bool,
    unique_pipeline_id: str,
) -> Tuple[Outputs, status.Status]:
    """The implementation of a single component runner.

    Returns a tuple of (outputs, status). If status is FAILURE, outputs
    is an empty dictionary.
    """

    task_resource_name = executor_input_utils.get_local_task_resource_name(
        component_name)
    task_root = executor_input_utils.construct_local_task_root(
        pipeline_root=pipeline_root,
        pipeline_resource_name=pipeline_resource_name,
        task_resource_name=task_resource_name,
    )
    executor_input = executor_input_utils.construct_executor_input(
        component_spec=component_spec,
        arguments=arguments,
        task_root=task_root,
        block_input_artifact=block_input_artifact,
    )

    container = executor_spec.container
    image = container.image
    full_command = list(container.command) + list(container.args)

    executor_input_dict = executor_input_utils.executor_input_to_dict(
        executor_input=executor_input,
        component_spec=component_spec,
    )
    full_command = placeholder_utils.replace_placeholders(
        full_command=full_command,
        executor_input_dict=executor_input_dict,
        pipeline_resource_name=pipeline_resource_name,
        task_resource_name=task_resource_name,
        pipeline_root=pipeline_root,
        unique_pipeline_id=unique_pipeline_id,
    )

    runner_type = type(runner)
    task_handler_map: Dict[
        local.LocalRunnerType, task_handler_interface.ITaskHandler] = {
            local.SubprocessRunner:
                subprocess_task_handler.SubprocessTaskHandler,
            local.DockerRunner:
                docker_task_handler.DockerTaskHandler,
        }
    TaskHandler = task_handler_map[runner_type]

    with logging_utils.local_logger_context():
        task_name_for_logs = logging_utils.format_task_name(task_resource_name)

        logging.info(f'Executing task {task_name_for_logs}')
        task_handler = TaskHandler(
            image=image,
            full_command=full_command,
            pipeline_root=pipeline_root,
            runner=runner,
        )

        # trailing newline helps visually separate subprocess logs
        logging.info(f'Streamed logs:\n')

        with logging_utils.indented_print():
            # subprocess logs printed here
            task_status = task_handler.run()

        if task_status == status.Status.SUCCESS:
            logging.info(
                f'Task {task_name_for_logs} finished with status {logging_utils.format_status(task_status)}'
            )

            outputs = executor_output_utils.get_outputs_for_task(
                executor_input=executor_input,
                component_spec=component_spec,
            )
            if outputs:
                output_string = [
                    f'Task {task_name_for_logs} outputs:',
                    *logging_utils.make_log_lines_for_outputs(outputs),
                ]
                logging.info('\n'.join(output_string))
            else:
                logging.info(f'Task {task_name_for_logs} has no outputs')

        elif task_status == status.Status.FAILURE:
            msg = f'Task {task_name_for_logs} finished with status {logging_utils.format_status(task_status)}'
            if raise_on_error:
                raise RuntimeError(msg)
            else:
                logging.error(msg)
            outputs = {}

        else:
            # for developers; user should never hit this
            raise ValueError(f'Got unknown status: {task_status}')
        logging_utils.print_horizontal_line()

        return outputs, task_status



# Copyright 2023 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from typing import Any, Dict, List

from kfp.local import config
from kfp.local import status
from kfp.local import task_handler_interface


class DockerTaskHandler(task_handler_interface.ITaskHandler):
    """The task handler corresponding to DockerRunner."""

    def __init__(
        self,
        image: str,
        full_command: List[str],
        pipeline_root: str,
        runner: config.DockerRunner,
    ) -> None:
        self.image = image
        self.full_command = full_command
        self.pipeline_root = pipeline_root
        self.runner = runner

    def get_volumes_to_mount(self) -> Dict[str, Any]:
        """Gets the volume configuration to mount the pipeline root to the
        container so that outputs can be obtained outside of the container."""
        if not os.path.isabs(self.pipeline_root):
            # defensive check. this is enforced by upstream code.
            # users should not hit this,
            raise ValueError(
                "'pipeline_root' should be an absolute path to correctly construct the volume mount specification."
            )
        return {self.pipeline_root: {'bind': self.pipeline_root, 'mode': 'rw'}}

    def run(self) -> status.Status:
        """Runs the Docker container and returns the status."""
        # nest docker import in case not available in user env so that
        # this module is runnable, even if not using DockerRunner
        import docker
        client = docker.from_env()
        try:
            volumes = self.get_volumes_to_mount()
            return_code = run_docker_container(
                client=client,
                image=self.image,
                command=self.full_command,
                volumes=volumes,
                **self.runner.container_run_args)
        finally:
            client.close()
        return status.Status.SUCCESS if return_code == 0 else status.Status.FAILURE


def add_latest_tag_if_not_present(image: str) -> str:
    if ':' not in image:
        image = f'{image}:latest'
    return image


def run_docker_container(client: 'docker.DockerClient', image: str,
                         command: List[str], volumes: Dict[str, Any],
                         **container_run_args) -> int:
    image = add_latest_tag_if_not_present(image=image)
    image_exists = any(
        image in (existing_image.tags + existing_image.attrs['RepoDigests'])
        for existing_image in client.images.list())
    if image_exists:
        print(f'Found image {image!r}\n')
    else:
        print(f'Pulling image {image!r}')
        client.images.pull(image)
        print('Image pull complete\n')
    container = client.containers.run(
        image=image,
        command=command,
        detach=True,
        stdout=True,
        stderr=True,
        volumes=volumes,
        **container_run_args)
    for line in container.logs(stream=True):
        # the inner logs should already have trailing \n
        # we do not need to add another
        print(line.decode(), end='')
    return container.wait()['StatusCode']



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Definition of PipelineChannel."""

import abc
import contextlib
import dataclasses
import json
import re
from typing import Dict, List, Optional, Union

from kfp.dsl.types import type_utils


@dataclasses.dataclass
class ConditionOperation:
    """Represents a condition expression to be used in condition control flow
    group.

    Attributes:
      operator: The operator of the condition.
      left_operand: The left operand.
      right_operand: The right operand.
      negate: Whether to negate the result of the binary operation.
    """
    operator: str
    left_operand: Union['PipelineParameterChannel', type_utils.PARAMETER_TYPES]
    right_operand: Union['PipelineParameterChannel', type_utils.PARAMETER_TYPES]
    negate: bool = False


# The string template used to generate the placeholder of a PipelineChannel.
_PIPELINE_CHANNEL_PLACEHOLDER_TEMPLATE = (
    '{{channel:task=%s;name=%s;type=%s;}}')
# The regex for parsing PipelineChannel placeholders from a string.
_PIPELINE_CHANNEL_PLACEHOLDER_REGEX = (
    r'{{channel:task=([\w\s_-]*);name=([\w\s_-]+);type=([\w\s{}":_-]*);}}')


class PipelineChannel(abc.ABC):
    """Represents a future value that is passed between pipeline components.

    A PipelineChannel object can be used as a pipeline function argument so that
    it will be a pipeline artifact or parameter that shows up in ML Pipelines
    system UI. It can also represent an intermediate value passed between
    components.

    Attributes:
        name: The name of the pipeline channel.
        channel_type: The type of the pipeline channel.
        task_name: The name of the task that produces the pipeline channel.
            None means it is not produced by any task, so if None, either user
            constructs it directly (for providing an immediate value), or it is
            a pipeline function argument.
        pattern: The serialized string regex pattern this pipeline channel
            created from.
    """

    @abc.abstractmethod
    def __init__(
        self,
        name: str,
        channel_type: Union[str, Dict],
        task_name: Optional[str] = None,
    ):
        """Initializes a PipelineChannel instance.

        Args:
            name: The name of the pipeline channel. The name will be sanitized
                to be k8s compatible.
            channel_type: The type of the pipeline channel.
            task_name: Optional; The name of the task that produces the pipeline
                channel. If provided, the task name will be sanitized to be k8s
                compatible.

        Raises:
            ValueError: If name or task_name contains invalid characters.
            ValueError: If both task_name and value are set.
        """
        valid_name_regex = r'^[A-Za-z][A-Za-z0-9\s_-]*$'
        if not re.match(valid_name_regex, name):
            raise ValueError(
                f'Only letters, numbers, spaces, "_", and "-" are allowed in the name. Must begin with a letter. Got name: {name}'
            )

        self.name = name
        self.channel_type = channel_type
        # ensure value is None even if empty string or empty list/dict
        # so that serialization and unserialization remain consistent
        # (i.e. None => '' => None)
        self.task_name = task_name or None
        from kfp.dsl import pipeline_context

        self.pipeline = pipeline_context.Pipeline.get_default_pipeline()

    @property
    def task(self) -> Union['PipelineTask', 'TasksGroup']:
        # TODO: migrate Collected to OneOfMixin style implementation,
        # then move this out of a property
        if self.task_name is None or self.pipeline is None:
            return None

        if self.task_name in self.pipeline.tasks:
            return self.pipeline.tasks[self.task_name]

        from kfp.compiler import compiler_utils
        all_groups = compiler_utils.get_all_groups(self.pipeline.groups[0])
        # pipeline hasn't exited, so it doesn't have a name
        all_groups_no_pipeline = all_groups[1:]
        group_name_to_group = {
            group.name: group for group in all_groups_no_pipeline
        }
        if self.task_name in group_name_to_group:
            return group_name_to_group[self.task_name]

        raise ValueError(
            f"PipelineChannel task name '{self.task_name}' not found in pipeline."
        )

    @property
    def full_name(self) -> str:
        """Unique name for the PipelineChannel."""
        return f'{self.task_name}-{self.name}' if self.task_name else self.name

    @property
    def pattern(self) -> str:
        """Unique pattern for the PipelineChannel."""
        return str(self)

    def __str__(self) -> str:
        """String representation of the PipelineChannel.

        The string representation is a string identifier so we can mix
        the PipelineChannel inline with other strings such as arguments.
        For example, we can support: ['echo %s' % param] as the
        container command and later a compiler can replace the
        placeholder '{{pipeline_channel:task=%s;name=%s;type=%s}}' with
        its own parameter identifier.
        """
        task_name = self.task_name or ''
        name = self.name
        channel_type = self.channel_type or ''
        if isinstance(channel_type, dict):
            channel_type = json.dumps(channel_type)
        return _PIPELINE_CHANNEL_PLACEHOLDER_TEMPLATE % (task_name, name,
                                                         channel_type)

    def __repr__(self) -> str:
        """Representation of the PipelineChannel.

        We make repr return the placeholder string so that if someone
        uses str()-based serialization of complex objects containing
        `PipelineChannel`, it works properly. (e.g. str([1, 2, 3,
        kfp.pipeline_channel.PipelineParameterChannel("aaa"), 4, 5, 6,]))
        """
        return str(self)

    def __hash__(self) -> int:
        """Returns the hash of a PipelineChannel."""
        return hash(self.pattern)

    def __eq__(self, other):
        return ConditionOperation('==', self, other)

    def __ne__(self, other):
        return ConditionOperation('!=', self, other)

    def __lt__(self, other):
        return ConditionOperation('<', self, other)

    def __le__(self, other):
        return ConditionOperation('<=', self, other)

    def __gt__(self, other):
        return ConditionOperation('>', self, other)

    def __ge__(self, other):
        return ConditionOperation('>=', self, other)


class PipelineParameterChannel(PipelineChannel):
    """Represents a pipeline parameter channel.

    Attributes:
      name: The name of the pipeline channel.
      channel_type: The type of the pipeline channel.
      task_name: The name of the task that produces the pipeline channel.
        None means it is not produced by any task, so if None, either user
        constructs it directly (for providing an immediate value), or it is a
        pipeline function argument.
      pattern: The serialized string regex pattern this pipeline channel created
        from.
      value: The actual value of the pipeline channel. If provided, the
        pipeline channel is "resolved" immediately.
    """

    def __init__(
        self,
        name: str,
        channel_type: Union[str, Dict],
        task_name: Optional[str] = None,
        value: Optional[type_utils.PARAMETER_TYPES] = None,
    ):
        """Initializes a PipelineArtifactChannel instance.

        Args:
          name: The name of the pipeline channel.
          channel_type: The type of the pipeline channel.
          task_name: Optional; The name of the task that produces the pipeline
            channel.
          value: Optional; The actual value of the pipeline channel.

        Raises:
          ValueError: If name or task_name contains invalid characters.
          ValueError: If both task_name and value are set.
          TypeError: If the channel type is not a parameter type.
        """
        if task_name and value:
            raise ValueError('task_name and value cannot be both set.')

        if not type_utils.is_parameter_type(channel_type):
            raise TypeError(f'{channel_type} is not a parameter type.')

        self.value = value

        super(PipelineParameterChannel, self).__init__(
            name=name,
            channel_type=channel_type,
            task_name=task_name,
        )


class PipelineArtifactChannel(PipelineChannel):
    """Represents a pipeline artifact channel.

    Attributes:
      name: The name of the pipeline channel.
      channel_type: The type of the pipeline channel.
      task_name: The name of the task that produces the pipeline channel.
        A pipeline artifact channel is always produced by some task.
      pattern: The serialized string regex pattern this pipeline channel created
        from.
    """

    def __init__(
        self,
        name: str,
        channel_type: Union[str, Dict],
        task_name: Optional[str],
        is_artifact_list: bool,
    ):
        """Initializes a PipelineArtifactChannel instance.

        Args:
            name: The name of the pipeline channel.
            channel_type: The type of the pipeline channel.
            task_name: Optional; the name of the task that produces the pipeline
                channel.
            is_artifact_list: True if `channel_type` represents a list of the artifact type.

        Raises:
            ValueError: If name or task_name contains invalid characters.
            TypeError: If the channel type is not an artifact type.
        """
        if type_utils.is_parameter_type(channel_type):
            raise TypeError(f'{channel_type} is not an artifact type.')

        self.is_artifact_list = is_artifact_list

        super(PipelineArtifactChannel, self).__init__(
            name=name,
            channel_type=channel_type,
            task_name=task_name,
        )


class OneOfMixin(PipelineChannel):
    """Shared functionality for OneOfParameter and OneOfAritfact."""

    def _set_condition_branches_group(
        self, channels: List[Union[PipelineParameterChannel,
                                   PipelineArtifactChannel]]
    ) -> None:
        # avoid circular import
        from kfp.dsl import tasks_group

        # .condition_branches_group could really be collapsed into just .task,
        # but we prefer keeping both for clarity in the rest of the compiler
        # code. When the code is logically related to a
        # condition_branches_group, it aids understanding to reference this
        # attribute name. When the code is trying to treat the OneOfMixin like
        # a typical PipelineChannel, it aids to reference task.
        self.condition_branches_group: tasks_group.ConditionBranches = channels[
            0].task.parent_task_group.parent_task_group

    def _make_oneof_name(self) -> str:
        # avoid circular imports
        from kfp.compiler import compiler_utils

        # This is a different type of "injected channel".
        # We know that this output will _always_ be a pipeline channel, so we
        # set the pipeline-channel-- prefix immediately (here).
        # In the downstream compiler logic, we get to treat this output like a
        # normal task output.
        return compiler_utils.additional_input_name_for_pipeline_channel(
            f'{self.condition_branches_group.name}-oneof-{self.condition_branches_group.get_oneof_id()}'
        )

    def _validate_channels(
        self,
        channels: List[Union[PipelineParameterChannel,
                             PipelineArtifactChannel]],
    ):
        self._validate_no_collected_channel(channels)
        self._validate_no_oneof_channel(channels)
        self._validate_no_mix_of_parameters_and_artifacts(channels)
        self._validate_has_else_group(self.condition_branches_group)

    def _validate_no_collected_channel(
        self, channels: List[Union[PipelineParameterChannel,
                                   PipelineArtifactChannel]]
    ) -> None:
        # avoid circular imports
        from kfp.dsl import for_loop
        if any(isinstance(channel, for_loop.Collected) for channel in channels):
            raise ValueError(
                f'dsl.{for_loop.Collected.__name__} cannot be used inside of dsl.{OneOf.__name__}.'
            )

    def _validate_no_oneof_channel(
        self, channels: List[Union[PipelineParameterChannel,
                                   PipelineArtifactChannel]]
    ) -> None:
        if any(isinstance(channel, OneOfMixin) for channel in channels):
            raise ValueError(
                f'dsl.{OneOf.__name__} cannot be used inside of another dsl.{OneOf.__name__}.'
            )

    def _validate_no_mix_of_parameters_and_artifacts(
        self, channels: List[Union[PipelineParameterChannel,
                                   PipelineArtifactChannel]]
    ) -> None:

        first_channel = channels[0]
        if isinstance(first_channel, PipelineParameterChannel):
            first_channel_type = PipelineParameterChannel
        else:
            first_channel_type = PipelineArtifactChannel

        for channel in channels:
            # if not all channels match the first channel's type, then there
            # is a mix of parameter and artifact channels
            if not isinstance(channel, first_channel_type):
                raise TypeError(
                    f'Task outputs passed to dsl.{OneOf.__name__} must be the same type. Found a mix of parameters and artifacts passed to dsl.{OneOf.__name__}.'
                )

    def _validate_has_else_group(
        self,
        parent_group: 'tasks_group.ConditionBranches',
    ) -> None:
        # avoid circular imports
        from kfp.dsl import tasks_group
        if not isinstance(parent_group.groups[-1], tasks_group.Else):
            raise ValueError(
                f'dsl.{OneOf.__name__} must include an output from a task in a dsl.{tasks_group.Else.__name__} group to ensure at least one output is available at runtime.'
            )

    def __str__(self):
        # supporting oneof in f-strings is technically feasible, but would
        # require somehow encoding all of the oneof channels into the
        # f-string
        # another way to do this would be to maintain a pipeline-level
        # map of PipelineChannels and encode a lookup key in the f-string
        # the combination of OneOf and an f-string is not common, so prefer
        # deferring implementation
        raise NotImplementedError(
            f'dsl.{OneOf.__name__} does not support string interpolation.')

    @property
    def pattern(self) -> str:
        # override self.pattern to avoid calling __str__, allowing us to block f-strings for now
        # this makes it OneOfMixin hashable for use in sets/dicts
        task_name = self.task_name or ''
        name = self.name
        channel_type = self.channel_type or ''
        if isinstance(channel_type, dict):
            channel_type = json.dumps(channel_type)
        return _PIPELINE_CHANNEL_PLACEHOLDER_TEMPLATE % (task_name, name,
                                                         channel_type)


# splitting out OneOf into subclasses significantly decreases the amount of
# branching in downstream compiler logic, since the
# isinstance(<some channel>, PipelineParameterChannel/PipelineArtifactChannel)
# checks continue to behave in desirable ways
class OneOfParameter(PipelineParameterChannel, OneOfMixin):
    """OneOf that results in a parameter channel for all downstream tasks."""

    def __init__(self, channels: List[PipelineParameterChannel]) -> None:
        self.channels = channels
        self._set_condition_branches_group(channels)
        super().__init__(
            name=self._make_oneof_name(),
            channel_type=channels[0].channel_type,
            task_name=None,
        )
        self.task_name = self.condition_branches_group.name
        self.channels = channels
        self._validate_channels(channels)
        self._validate_same_kfp_type(channels)

    def _validate_same_kfp_type(
            self, channels: List[PipelineParameterChannel]) -> None:
        expected_type = channels[0].channel_type
        for i, channel in enumerate(channels[1:], start=1):
            if channel.channel_type != expected_type:
                raise TypeError(
                    f'Task outputs passed to dsl.{OneOf.__name__} must be the same type. Got two channels with different types: {expected_type} at index 0 and {channel.channel_type} at index {i}.'
                )


class OneOfArtifact(PipelineArtifactChannel, OneOfMixin):
    """OneOf that results in an artifact channel for all downstream tasks."""

    def __init__(self, channels: List[PipelineArtifactChannel]) -> None:
        self.channels = channels
        self._set_condition_branches_group(channels)
        super().__init__(
            name=self._make_oneof_name(),
            channel_type=channels[0].channel_type,
            task_name=None,
            is_artifact_list=channels[0].is_artifact_list,
        )
        self.task_name = self.condition_branches_group.name
        self._validate_channels(channels)
        self._validate_same_kfp_type(channels)

    def _validate_same_kfp_type(
            self, channels: List[PipelineArtifactChannel]) -> None:
        # Unlike for component interface type checking where anything is
        # passable to Artifact, we should require the output artifacts for a
        # OneOf to be the same. This reduces the complexity/ambiguity for the
        # user of the actual type checking logic. What should the type checking
        # behavior be if the OneOf surfaces an Artifact and a Dataset? We can
        # always loosen backward compatibly in the future, so prefer starting
        # conservatively.
        expected_type = channels[0].channel_type
        expected_is_list = channels[0].is_artifact_list
        for i, channel in enumerate(channels[1:], start=1):
            if channel.channel_type != expected_type or channel.is_artifact_list != expected_is_list:
                raise TypeError(
                    f'Task outputs passed to dsl.{OneOf.__name__} must be the same type. Got two channels with different types: {expected_type} at index 0 and {channel.channel_type} at index {i}.'
                )


class OneOf:
    """For collecting mutually exclusive outputs from conditional branches into
    a single pipeline channel.

    Args:
        channels: The channels to collect into a OneOf. Must be of the same type.

    Example:
      ::

        @dsl.pipeline
        def flip_coin_pipeline() -> str:
            flip_coin_task = flip_coin()
            with dsl.If(flip_coin_task.output == 'heads'):
                print_task_1 = print_and_return(text='Got heads!')
            with dsl.Else():
                print_task_2 = print_and_return(text='Got tails!')

            # use the output from the branch that gets executed
            oneof = dsl.OneOf(print_task_1.output, print_task_2.output)

            # consume it
            print_and_return(text=oneof)

            # return it
            return oneof
    """

    def __new__(
        cls, *channels: Union[PipelineParameterChannel, PipelineArtifactChannel]
    ) -> Union[OneOfParameter, OneOfArtifact]:
        first_channel = channels[0]
        if isinstance(first_channel, PipelineParameterChannel):
            return OneOfParameter(channels=list(channels))
        elif isinstance(first_channel, PipelineArtifactChannel):
            return OneOfArtifact(channels=list(channels))
        else:
            raise ValueError(
                f'Got unknown input to dsl.{OneOf.__name__} with type {type(first_channel)}.'
            )


def create_pipeline_channel(
    name: str,
    channel_type: Union[str, Dict],
    task_name: Optional[str] = None,
    value: Optional[type_utils.PARAMETER_TYPES] = None,
    is_artifact_list: bool = False,
) -> PipelineChannel:
    """Creates a PipelineChannel object.

    Args:
        name: The name of the channel.
        channel_type: The type of the channel, which decides whether it is an
            PipelineParameterChannel or PipelineArtifactChannel
        task_name: Optional; the task that produced the channel.
        value: Optional; the realized value for a channel.

    Returns:
        A PipelineParameterChannel or PipelineArtifactChannel object.
    """
    if type_utils.is_parameter_type(channel_type):
        return PipelineParameterChannel(
            name=name,
            channel_type=channel_type,
            task_name=task_name,
            value=value,
        )
    else:
        return PipelineArtifactChannel(
            name=name,
            channel_type=channel_type,
            task_name=task_name,
            is_artifact_list=is_artifact_list,
        )


def extract_pipeline_channels_from_string(
        payload: str) -> List[PipelineChannel]:
    """Extracts a list of PipelineChannel instances from the payload string.

    Note: this function removes all duplicate matches.

    Args:
      payload: A string that may contain serialized PipelineChannels.

    Returns:
      A list of PipelineChannels found from the payload.
    """
    matches = re.findall(_PIPELINE_CHANNEL_PLACEHOLDER_REGEX, payload)
    unique_channels = set()
    for match in matches:
        task_name, name, channel_type = match

        # channel_type could be either a string (e.g. "Integer") or a dictionary
        # (e.g.: {"custom_type": {"custom_property": "some_value"}}).
        # Try loading it into dictionary, if failed, it means channel_type is a
        # string.
        with contextlib.suppress(json.JSONDecodeError):
            channel_type = json.loads(channel_type)

        if type_utils.is_parameter_type(channel_type):
            pipeline_channel = PipelineParameterChannel(
                name=name,
                channel_type=channel_type,
                task_name=task_name,
            )
        else:
            pipeline_channel = PipelineArtifactChannel(
                name=name,
                channel_type=channel_type,
                task_name=task_name,
                # currently no support for getting the index from a list of artifacts (e.g., my_datasets[0].uri), so this will always be False until accessing a single artifact element is supported
                is_artifact_list=False,
            )
        unique_channels.add(pipeline_channel)

    return list(unique_channels)


def extract_pipeline_channels_from_any(
    payload: Union[PipelineChannel, str, int, float, bool, list, tuple, dict]
) -> List[PipelineChannel]:
    """Recursively extract PipelineChannels from any object or list of objects.

    Args:
      payload: An object that contains serialized PipelineChannels or k8
        definition objects.

    Returns:
      A list of PipelineChannels found from the payload.
    """
    if not payload:
        return []

    if isinstance(payload, PipelineChannel):
        return [payload]

    if isinstance(payload, str):
        return list(set(extract_pipeline_channels_from_string(payload)))

    if isinstance(payload, (list, tuple)):
        pipeline_channels = []
        for item in payload:
            pipeline_channels += extract_pipeline_channels_from_any(item)
        return list(set(pipeline_channels))

    if isinstance(payload, dict):
        pipeline_channels = []
        for key, value in payload.items():
            pipeline_channels += extract_pipeline_channels_from_any(key)
            pipeline_channels += extract_pipeline_channels_from_any(value)
        return list(set(pipeline_channels))

    # TODO(chensun): extract PipelineChannel from v2 container spec?

    return []
